import math
import pytest
from dtaidistance import dtw, util_numpy

numpyonly = pytest.mark.skipif("util_numpy.test_without_numpy()")


@numpyonly
def test_penalty_cyclicalshift():
    with util_numpy.test_uses_numpy() as np:
        np.set_printoptions(precision=2, linewidth=120)
        s1 = np.array([0., 1, 2, 1, 0, 1, 2, 1, 0, 1, 2, 1, 0])
        s2 = np.array([2., 1, 0, 1, 2, 1, 0, 1, 2, 1, 0, 1, 2])
        # plt.plot(s1)
        # plt.plot(s2)
        # plt.show(block=True)
        d1 = dtw.distance(s1, s2)
        d2 = dtw.distance(s1, s2, penalty=1)
        assert d1 == pytest.approx(math.sqrt(10))
        assert d2 == pytest.approx(math.sqrt(14))


if __name__ == "__main__":
    test_penalty_cyclicalshift()



{%- set s = " " %}
{%- if "ndim" in suffix %}
{%- set ndim = "ndim" %}
{%- else %}
{%- set ndim = "1" %}
{%- endif %}
def dba{{suffix}}(cur,{{s}}
        {%- if "ndim" in suffix -%}
        {{seq_tpy}}[:, :] c, unsigned char[:] mask, int nb_prob_samples, int ndim,{{s}}
        {%- else -%}
        {{seq_tpy}}[:] c, unsigned char[:] mask, int nb_prob_samples,{{s}}
        {%- endif -%}
        **kwargs):
    {%- if "ndim" in suffix %}
    cdef {{seq_tpy}} *c_ptr = &c[0, 0];
    {%- else %}
    cdef {{seq_tpy}} *c_ptr = &c[0];
    {%- endif %}
    cdef unsigned char *mask_ptr = &mask[0];
    settings = DTWSettings(**kwargs)
    dba_inner(cur, c_ptr, len(c), mask_ptr, nb_prob_samples, {{ndim}}, settings)
    return c



{%- set s = " " %}
def warping_path{{suffix}}(
        {%- if "ndim" in suffix -%}
        {{seq_tpy}}[:, :] s1, {{seq_tpy}}[:, :] s2, int ndim=1,{{s}}
        {%- else -%}
        {{seq_tpy}}[:] s1, {{seq_tpy}}[:] s2,{{s}}
        {%- endif -%}
        **kwargs):
    # Assumes C contiguous
    cdef Py_ssize_t path_length;
    settings = DTWSettings(**kwargs)
    cdef Py_ssize_t *i1 = <Py_ssize_t *> PyMem_Malloc((len(s1) + len(s2)) * sizeof(Py_ssize_t))
    if not i1:
        raise MemoryError()
    cdef Py_ssize_t *i2 = <Py_ssize_t *> PyMem_Malloc((len(s1) + len(s2)) * sizeof(Py_ssize_t))
    if not i2:
        raise MemoryError()
    try:
        {%- if "ndim" in suffix %}
        path_length = dtaidistancec_dtw.warping_path_ndim(&s1[0, 0], len(s1), &s2[0, 0], len(s2), i1, i2, ndim, &settings._settings)
        {%- else %}
        path_length = dtaidistancec_dtw.warping_path(&s1[0], len(s1), &s2[0], len(s2), i1, i2, &settings._settings)
        {%- endif %}
        path = []
        for i in range(path_length):
            path.append((i1[i], i2[i]))
        path.reverse()
    finally:
        PyMem_Free(i1)
        PyMem_Free(i2)
    return path



{% macro select_c_fn(wps_view) -%}
    {%- if "affinity" in suffix %}
    {%- if "ndim" in suffix %}
    d = dtaidistancec_dtw.dtw_warping_paths_affinity_ndim(&{{wps_view}}[0,0], &s1[0,0], len(s1), &s2[0,0], len(s2),
                                                          True, False, psi_neg, only_triu, ndim,
                                                          gamma, tau, delta, delta_factor,
                                                          &settings._settings)
    {%- else %}
    d = dtaidistancec_dtw.dtw_warping_paths_affinity(&{{wps_view}}[0,0], &s1[0], len(s1), &s2[0], len(s2),
                                                     True, False, psi_neg, only_triu,
                                                     gamma, tau, delta, delta_factor,
                                                     &settings._settings)
    {%- endif %}
    {%- else %}
    {%- if "ndim" in suffix %}
    d = dtaidistancec_dtw.dtw_warping_paths_ndim(&{{wps_view}}[0,0], &s1[0,0], len(s1), &s2[0,0], len(s2),
                                                 True, True, psi_neg, ndim, &settings._settings)
    {%- else %}
    d = dtaidistancec_dtw.dtw_warping_paths(&{{wps_view}}[0,0], &s1[0], len(s1), &s2[0], len(s2),
                                            True, True, psi_neg, &settings._settings)
    {%- endif %}
    {%- endif %}
{%- endmacro -%}
{%- set s = " " -%}


def warping_paths{{ suffix }}(
        {%- if "ndim" in suffix -%}
        {{seq_tpy}}[:, :] dtw, {{seq_tpy}}[:, :] s1, {{seq_tpy}}[:, :] s2,{{s}}
        {%- else -%}
        {{seq_tpy}}[:, :] dtw, {{seq_tpy}}[:] s1, {{seq_tpy}}[:] s2,{{s}}
        {%- endif -%}
        {%- if "affinity" in suffix %}
        bint only_triu, {{seq_tpy}} gamma, {{seq_tpy}} tau, {{seq_tpy}} delta, {{seq_tpy}} delta_factor,
        {%- endif %}
        bint psi_neg=False, **kwargs):
    {%- if "ndim" in suffix %}
    ndim = s1.shape[1]
    if s1.shape[1] != s2.shape[1]:
        raise Exception("Dimension of sequence entries needs to be the same: {} != {}".format(s1.shape[1], s2.shape[1]))
    {%- endif %}
    # Assumes C contiguous
    settings = DTWSettings(**kwargs)
    dtw_length = dtw.shape[0] * dtw.shape[1]
    req_length = dtaidistancec_dtw.dtw_settings_wps_length(len(s1), len(s2), &settings._settings)
    req_width = dtaidistancec_dtw.dtw_settings_wps_width(len(s1), len(s2), &settings._settings)
    shape = (1, req_length)
    if req_length == dtw_length and req_width == dtw.shape[1]:
        # No compact WPS array is required
        wps = dtw
    else:
        try:
            # Use cython.view.array to avoid numpy dependency
            wps = cvarray(shape=shape, itemsize=sizeof({{seq_tpy}}), format="d")
        except MemoryError as exc:
            print("Cannot allocate memory for warping paths matrix. Trying " + str(shape) + ".")
            raise exc
    cdef {{seq_tpy}} [:, :] wps_view = wps
    cdef {{seq_tpy}} d
    {{ select_c_fn("wps_view")}}
    if not (req_length == dtw_length and req_width == dtw.shape[1]):
        {%- if "affinity" in suffix %}
        dtaidistancec_dtw.dtw_expand_wps_affinity(&wps_view[0,0], &dtw[0, 0], len(s1), len(s2), &settings._settings)
        {%- else %}
        dtaidistancec_dtw.dtw_expand_wps(&wps_view[0,0], &dtw[0, 0], len(s1), len(s2), &settings._settings)
        {%- endif %}
    return d


def warping_paths_compact{{ suffix }}(
        {%- if "ndim" in suffix -%}
        {{seq_tpy}}[:, :] dtw, {{seq_tpy}}[:, :] s1, {{seq_tpy}}[:, :] s2,{{s}}
        {%- else -%}
        {{seq_tpy}}[:, :] dtw, {{seq_tpy}}[:] s1, {{seq_tpy}}[:] s2,{{s}}
        {%- endif -%}
        {%- if "affinity" in suffix %}
        bint only_triu, {{seq_tpy}} gamma, {{seq_tpy}} tau, {{seq_tpy}} delta, {{seq_tpy}} delta_factor,
        {%- endif %}
        bint psi_neg=False, **kwargs):
    {%- if "ndim" in suffix %}
    if s1.shape[1] != s2.shape[1]:
        raise Exception("Dimension of sequence entries needs to be the same: {} != {}".format(s1.shape[1], s2.shape[1]))
    ndim = s1.shape[1]
    {%- endif %}
    # Assumes C contiguous
    settings = DTWSettings(**kwargs)
    cdef {{seq_tpy}} d
    {{ select_c_fn("dtw") }}
    return d



# -*- coding: UTF-8 -*-
"""
dtaidistance.util
~~~~~~~~~~~~~~~~~

Utility functions for DTAIDistance.

:author: Wannes Meert
:copyright: Copyright 2017-2022 KU Leuven, DTAI Research Group.
:license: Apache License, Version 2.0, see LICENSE for details.

"""
import os
import sys
import csv
import logging
from array import array
from pathlib import Path
import tempfile


logger = logging.getLogger("be.kuleuven.dtai.distance")


try:
    import numpy as np
except ImportError:
    np = None

try:
    from . import dtw_cc
except ImportError:
    dtw_cc = None

try:
    from . import dtw_cc_omp
except ImportError:
    dtw_cc_omp = None

try:
    from . import dtw_cc_numpy
except ImportError:
    dtw_cc_numpy = None
except ValueError as exc:
    logger.warning('Warning: loading library to link with numpy returned an error')
    logger.warning(exc)
    dtw_cc_numpy = None


dtaidistance_dir = os.path.abspath(os.path.dirname(__file__))


def try_import_c(verbose=False):
    is_complete = True
    msgs = []
    global dtw_cc
    global dtw_cc_omp
    global dtw_cc_numpy
    try:
        from . import dtw_cc
    except Exception as exc:
        print('Cannot import C-based library (dtw_cc)')
        msgs.append('Cannot import C-based library (dtw_cc)')
        msgs.append(str(exc))
        dtw_cc = None
        is_complete = False
    try:
        from . import dtw_cc_omp
    except Exception as exc:
        print('Cannot import OMP-based library (dtw_cc_omp)')
        msgs.append('Cannot import OMP-based library (dtw_cc_omp)')
        msgs.append(str(exc))
        dtw_cc_omp = None
        is_complete = False
    try:
        from . import dtw_cc_numpy
    except Exception as exc:
        print('Cannot import Numpy-based library (dtw_cc_numpy)')
        msgs.append('Cannot import Numpy-based library (dtw_cc_numpy)')
        msgs.append(str(exc))
        dtw_cc_numpy = None
        is_complete = False
    try:
        import numpy
        msgs.append('Numpy version: {}'.format(numpy.__version__))
    except Exception as exc:
        print('Cannot import Numpy (optional dependency)')
        msgs.append('Cannot import Numpy (optional dependency)')
        msgs.append(str(exc))
    try:
        import matplotlib
        msgs.append('Matplotlib version: {}'.format(matplotlib.__version__))
    except Exception as exc:
        print('Cannot import Matplotlib (optional dependency)')
        msgs.append('Cannot import Matplotlib (optional dependency)')
        msgs.append(str(exc))
    try:
        import scipy
        msgs.append('Scipy version: {}'.format(scipy.__version__))
    except Exception as exc:
        print('Cannot import SciPy (optional dependency)')
        msgs.append('Cannot import SciPy (optional dependency)')
        msgs.append(str(exc))
    if not is_complete:
        print('\nNot all libraries are available in your installation. ')
        print('You can rerun the compilation from source or pip install in verbose mode:\n'
              'pip install -vvv --upgrade --force-reinstall --no-deps --no-binary dtaidistance dtaidistance')
        print('In case you need to use an older version of numpy, compile against your current installation:\n'
              'pip install -vvv --upgrade --force-reinstall --no-deps --no-build-isolation '
              '--no-binary dtaidistance dtaidistance')
        print('\nShare the following information when submitting a bug report:')
    elif verbose:
        print('All ok ...')
    if not is_complete or verbose:
        print('== Packages ==')
        for msg in msgs:
            print('- {}'.format(msg))
        print('== System information ==')
        import sys
        print(sys.implementation)
        print('== Compilation information ==')
        try:
            import pkgutil
            logtxt = pkgutil.get_data(__name__, "compilation.log")
            print(logtxt.decode())
        except Exception as exc:
            print('Could not read compilation.log')
            print(exc)
        print('')
        print('==')
    return is_complete


def prepare_directory(directory=None):
    """Prepare the given directory, create it if necessary.
    If no directory is given, a new directory will be created in the system's temp directory.
    """
    if directory is not None:
        directory = Path(directory)
        if not directory.exists():
            directory.mkdir(parents=True)
        logger.debug("Using directory: {}".format(directory))
        return Path(directory)
    directory = tempfile.mkdtemp(prefix="dtaidistance_")
    logger.debug("Using directory: {}".format(directory))
    return Path(directory)


def read_substitution_matrix(file):
    """Read substitution matrix from file.

    Comments starting with # and newlines are allowed anywhere
    in the file.
    
    :return: A dictionary mapping tuples of symbols to their weight.
    """

    def strip_comments(reader):
        for line in reader:
            if not line.rstrip() or line[0] == '#':
                continue
            yield line.rstrip()

    matrix = dict()
    with open(file) as f:
        reader = csv.reader(strip_comments(f), delimiter=" ", skipinitialspace=True)
        line = next(reader)
        idx = {i: symbol for i, symbol in enumerate(line)}
        for line in reader:
            symbol = line[0]
            for j, value in enumerate(line[1:]):
                matrix[(idx[j], symbol)] = float(value)
    return matrix


class SeriesContainer:
    def __init__(self, series, support_ndim=True):
        """Container for a list of series.

        This wrapper class knows how to deal with multiple types of datastructures to represent
        a sequence of sequences.

        For 1 dimensional time series (e.g. [1,2,3,4]):
        - List[array.array]
        - List[numpy.array] (array is 1 dimensional)
        - List[List]
        - numpy.array (2 dimensional)
        - numpy.matrix

        For n-dimensional time series (e.g. [[1,2],[3,4],[5,6]]):
        - List[numpy.array] (array is 2 dimensional)
        - List[List[List]]
        - numpy.array (3 dimensional)

        When using the C-based extensions, the data is automatically verified and converted.
        """
        self.support_ndim = support_ndim
        # Always detect the dimensionality of the time series, even if support_ndim is false
        self.detected_ndim = False
        if isinstance(series, SeriesContainer):
            self.series = series.series
        elif np is not None and isinstance(series, np.ndarray):
            # A np.matrix always returns a 2D array, also if you select one row (to be consistent
            # and always be a matrix datastructure). The methods in this toolbox expect a
            # 1D array thus we need to convert to a 1D or 2D array. This is taken care by asarray
            self.series = np.asarray(series, order="C")
            if self.series.ndim > 2:
                if not self.support_ndim:
                    raise Exception('N-dimensional series are not supported '
                                    '(series.ndim = {}) > 2'.format(self.series.ndim))
                self.detected_ndim = len(self.series[0, 0])
            else:
                self.detected_ndim = 1
        elif type(series) in [set, tuple, list]:
            self.series = list(series)
            if np is not None and isinstance(self.series[0], np.ndarray):
                if self.series[0].ndim > 1:
                    if not self.support_ndim:
                        raise Exception('N-dimensional series are not supported '
                                        '(series[0].ndim = {}) > 1'.format(self.series[0].ndim))
                    self.detected_ndim = len(self.series[0][0])
                else:
                    self.detected_ndim = 1
            elif type(series[0]) in [list, tuple]:
                if type(series[0][0]) in [list, tuple]:
                    self.detected_ndim = len(self.series[0][0])
                else:
                    self.detected_ndim = 1
        else:
            self.series = series

    def c_data_compat(self):
        """Return a datastructure that the C-component knows how to handle.
        The method tries to avoid copying or reallocating memory.

        :return: Either a list of buffers or a two-dimensional buffer. The
            buffers are guaranteed to be C-contiguous and can thus be used
            as regular pointer-based arrays in C.
        """
        if dtw_cc is None:
            raise Exception('C library not loaded')
        if type(self.series) == list:
            for i in range(len(self.series)):
                serie = self.series[i]
                if np is not None and isinstance(serie, np.ndarray):
                    if not self.support_ndim and serie.ndim != 1:
                        raise Exception('N-dimensional arrays are not supported (serie.ndim = {})'.format(serie.ndim))
                    if not serie.flags.c_contiguous:
                        serie = np.asarray(serie, order="C")
                        self.series[i] = serie
                elif isinstance(serie, array):
                    pass
                else:
                    raise Exception(
                        "Type of series not supported, "
                        "expected numpy.array or array.array but got {}".format(
                            type(serie)
                        )
                    )
            return dtw_cc.dtw_series_from_data(self.series)
        elif np is not None and isinstance(self.series, np.ndarray):
            if not self.series.flags.c_contiguous:
                logger.warning("Numpy array not C contiguous, copying data.")
                self.series = self.series.copy(order="C")
            if not self.support_ndim and self.series.ndim > 2:
                raise Exception(f'N-dimensional series are not supported (series.ndim = {self.series.ndim})')
            if dtw_cc_numpy is None:
                logger.warning("DTAIDistance C-extension for Numpy is not available. Proceeding anyway.")
                return dtw_cc.dtw_series_from_data(self.series)
            elif len(self.series.shape) == 3:
                return dtw_cc_numpy.dtw_series_from_numpy_ndim(self.series)
            else:
                return dtw_cc_numpy.dtw_series_from_numpy(self.series)
        return dtw_cc.dtw_series_from_data(self.series)

    def get_max_y(self):
        max_y = 0
        if isinstance(self.series, np.ndarray) and len(self.series.shape) == 2:
            max_y = max(np.max(self.series), abs(np.min(self.series)))
        else:
            for serie in self.series:
                max_y = max(max_y, np.max(serie), abs(np.min(serie)))
        return max_y

    def get_max_min_y(self):
        min_y, max_y = 0, 0
        if isinstance(self.series, np.ndarray) and len(self.series.shape) == 2:
            min_y = np.min(self.series)
            max_y = np.max(self.series)
        else:
            for serie in self.series:
                min_y = min(min_y, np.min(serie))
                max_y = max(max_y, np.max(serie))
        return min_y, max_y

    def get_max_length(self):
        max_length = 0
        if isinstance(self.series, np.ndarray) and len(self.series.shape) == 2:
            max_length = self.series.shape[1]
        else:
            for serie in self.series:
                max_length = max(max_length, len(serie))
        return max_length

    def get_avg_length(self):
        max_length = 0
        if isinstance(self.series, np.ndarray) and len(self.series.shape) == 2:
            max_length = self.series.shape[1]
        else:
            for serie in self.series:
                max_length += len(serie)
            max_length /= len(self.series)
        return max_length

    def __getitem__(self, item):
        return self.series[item]

    def __len__(self):
        return len(self.series)

    def __str__(self):
        return "SeriesContainer:\n{}".format(self.series)

    @staticmethod
    def wrap(series, support_ndim=True):
        if isinstance(series, SeriesContainer):
            series.support_ndim = support_ndim
            return series
        return SeriesContainer(series, support_ndim=support_ndim)


def recompile():
    import subprocess as sp

    sp.run([sys.executable, "setup.py", "build_ext", "--inplace"], cwd=dtaidistance_dir)


def argmin(a):
    imin, vmin = 0, float("inf")
    for i, v in enumerate(a):
        if v < vmin:
            imin, vmin = i, v
    return imin


def argmax(a):
    imax, vmax = 0, float("-inf")
    for i, v in enumerate(a):
        if v > vmax:
            imax, vmax = i, v
    return imax

